#ifndef DECLARES_H
#define DECLARES_H
#endif

//#define TRACE_CONSTRUCTORS

// Include files

#pragma pack(4)
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#include <io.h>
#include <fcntl.H>
#include <SYS\STAT.H>
#include <ERRNO.H>

#include <math.h>
double PI = 3.1415926536;

#include "linkedlist.h"
#include "engine.h"

#include <winsock2.h>


// Objects

class PacketStruct;		// Packet
class PostlogNode;		// Postlog node
class Mutex;			// Safe multi-threading made easy
class BotSession;		// Socket implementation


// SubSpace Protocol-specific address descriptor

struct INADDR
{
	WORD Family;		// Socket family (Always INET_ADDR)
	WORD Port;			// Socket subclassing
	DWORD IP;			// Host address
	BYTE Padding[8];	// Unused space

	INADDR()
	{
		memset(this, 0, 16);

		// Write template for a SubSpace Service address
		Family = 2;
	}

	bool operator==(INADDR Other)
	{
		return ((Other.IP == IP) && (Other.Port == Port));
	}

	INADDR(DWORD NewIP, WORD NewPort)
	{
		memset(this, 0, 16);

		// Write a SubSpace Service address
		Family	= 2;
		IP		= NewIP;
		Port	= htons(NewPort);
	}

	char * GetStringIP()
	{
		in_addr in;
		in.S_un.S_addr = IP;

		return inet_ntoa(in);
	}
};


char * WSAGetErrorString(int ErrorCode)
{
	switch (ErrorCode)
	{
		case WSANOTINITIALISED:			return "WSANOTINITIALISED";
		case WSAENETDOWN:				return "WSAENETDOWN";
		case WSAEINPROGRESS:			return "WSAEINPROGRESS";
		case WSA_NOT_ENOUGH_MEMORY:		return "WSA_NOT_ENOUGH_MEMORY";
		case WSA_INVALID_HANDLE:		return "WSA_INVALID_HANDLE";
		case WSA_INVALID_PARAMETER:		return "WSA_INVALID_PARAMETER";
		case WSAEFAULT:					return "WSAEFAULT";
		case WSAEINTR:					return "WSAEINTR";
		case WSAEINVAL:					return "WSAEINVAL";
		case WSAEISCONN:				return "WSAEISCONN";
		case WSAENETRESET:				return "WSAENETRESET";
		case WSAENOTSOCK:				return "WSAENOTSOCK";
		case WSAEOPNOTSUPP:				return "WSAEOPNOTSUPP";
		case WSAESHUTDOWN:				return "WSAESHUTDOWN";
		case WSAEWOULDBLOCK:			return "WSAEWOULDBLOCK";
		case WSAEMSGSIZE:				return "WSAEMSGSIZE";
		case WSAETIMEDOUT:				return "WSAETIMEDOUT";
		case WSAECONNRESET:				return "WSAECONNRESET";
		case WSAENOTCONN:				return "WSAENOTCONN";
		case WSAEDISCON:				return "WSAEDISCON";
		case WSA_IO_PENDING:			return "WSA_IO_PENDING";
		case WSA_OPERATION_ABORTED:		return "WSA_OPERATION_ABORTED";
		default:						return "Unknown error";
	};
}


// System registry access

DWORD GetSettingDWORD(HKEY BaseKey, const char * Path, const char * Value)
{
	// Open the key
	DWORD key;			// Handle to a session with a registry key

	RegOpenKey((HKEY)BaseKey, Path, (HKEY*)&key);

	// Get a value
	DWORD buffer;		// Results of transaction will go here
	DWORD buflen = 4;	// Length of the buffer is 4, naturally
	DWORD type;			// Type will contain type of data transfered

	RegQueryValueEx((HKEY)key, Value, NULL, &type, (BYTE*)&buffer, &buflen);

	// Close the key
	RegCloseKey((HKEY)key);

	// Check data for validity
	if (type == REG_DWORD)
		return buffer;
	else
		return 0;			// 0 = error
}

void SetSettingDWORD(HKEY BaseKey, const char * Path, const char * Value, DWORD buffer)
{
	// Open the key
	DWORD key;			// Handle to a session with a registry key

	RegCreateKey((HKEY)BaseKey, Path, (HKEY*)&key);

	// Get a value
	RegSetValueEx((HKEY)key, Value, 0, REG_DWORD, (BYTE*)&buffer, 4);

	// Close the key
	RegCloseKey((HKEY)key);
}


// Game physics


/* -- How do I get burst shrap Thrust X & Y?
 * Given: BurstSpeed, TotalShrapnel, (ShrapnelIndex)
 * Calculations:
 * 1) (ShrapnelIndex / TotalShrapnel) * 2 * pi -> THETA
 * 2) BurstSpeed * sin(THETA) -> ThrustX
 * 3) BurstSpeed * cos(THETA) -> ThrustY
 */


// Wrapper for raw ship type numbers

enum ShipClasses
{
	ssWarbird = 0,
	ssJavelin,
	ssSpider,
	ssLeviathan,
	ssTerrier,
	ssWeasel,
	ssLancaster,
	ssShark,
	ssSpectator = 8
};


// Wrapper for access level numbers

enum AccessLevels
{
	ssPlayer,
	ssModerator,
	ssSuperModerator,
	ssSysOp
};


// Player entering constants

enum Login_Params
{
	NO_PROBLEMO		= 0,
	NEW_USER		= 1,
	BAD_PW			= 2,
	IP_BLOCK		= 3,
	ID_BLOCK		= 4,
	PERMISSION_ZONE	= 5,
	INVALID_NAME	= 11,
	OBSCENE_NAME	= 12,
	NO_SCORES		= 13,
	SERVER_BUSY		= 14,
	ADVANCED_ONLY	= 15,
	NAME_MISSING	= 16
};


// Chat parameters

enum Chat_Modes
{
    MSG_Arena,		// Same for *arena, *zone, **
    MSG_Unknown1,
    MSG_Public,
    MSG_Team,
    MSG_TeamPrivate,
    MSG_Private,
    MSG_Unknown6,
    MSG_RemotePrivate,
    MSG_ServerError,
    MSG_Channel
};


// Item info concatenated to the position packets C2S

struct ItemInfo
{
	unsigned Shields	: 1;
	unsigned Super		: 1;
	unsigned Burst		: 4;
	unsigned Repel		: 4;
	unsigned Thor		: 4;
	unsigned Wall		: 4;
	unsigned Decoy		: 4;
	unsigned Rocket		: 4;
	unsigned Portal		: 4;
	unsigned Pack		: 2;

	ItemInfo()
	{
	}

	ItemInfo(ItemInfo const &Items)
	{
		memcpy(this, &Items, 4);
	}
};

ItemInfo MakeItemInfo(bool Shields, bool Super, BYTE Burst, BYTE Repel, BYTE Portal, BYTE Decoy, BYTE Thor, BYTE Wall, BYTE Rocket)
{
	ItemInfo i;
	i.Shields	= Shields;
	i.Super		= Super;
	i.Burst		= Burst;
	i.Repel		= Repel;
	i.Portal	= Portal;
	i.Decoy		= Decoy;
	i.Thor		= Thor;
	i.Wall		= Wall;
	i.Rocket	= Rocket;
	i.Pack		= 3;
	return i;
}


// Weapon info concatenated to the weapons packets S2C

struct WeaponInfo
{
	unsigned Type		: 4;
	unsigned UNKNOWN0	: 1;
	unsigned Level		: 2;
	unsigned Bouncing	: 1;
	unsigned EMP		: 1;
	unsigned UNKNOWN1	: 1;
	unsigned ShrapCount	: 5;
	unsigned FireType	: 1;	// Bombs -> Mines, Bullets -> Multifire

	WeaponInfo()
	{
	}

	WeaponInfo(WeaponInfo const &Weapon)
	{
		memcpy(this, &Weapon, 2);
	}
};

WeaponInfo MakeWeaponInfo(BYTE nType, BYTE nLevel, BYTE nBouncing, BYTE nEMP, BYTE nShrapCount, BYTE nAlternateFireType, BYTE nU0, BYTE nU1)
{
	WeaponInfo w;
	w.Type		= nType;
	w.UNKNOWN0	= nU0;
	w.Level		= nLevel;
	w.Bouncing	= nBouncing;
	w.EMP		= nEMP;
	w.UNKNOWN1	= nU1;
	w.ShrapCount= nShrapCount;
	w.FireType	= nAlternateFireType;
	return w;
}


// Weapon types for WeaponInfo.Type

enum Projectile_Types
{
	// Seen "in the wild"
	PROJ_None,
	PROJ_Nothing = 0,
	PROJ_Bullet,
	PROJ_BBullet,
	PROJ_Bomb,
	PROJ_PBomb,
	PROJ_Repel,
	PROJ_Decoy,
	PROJ_Burst,
	PROJ_Thor,

	// Internal to the bot
	PROJ_InactiveBullet,
	PROJ_Shrapnel,
};


// Encapsulation of the player position packet parameters

struct PositionInfo
{
	unsigned Stealth		: 1;
	unsigned Cloaked		: 1;
	unsigned XRadar			: 1;
	unsigned Antiwarp		: 1;
	unsigned WarpFlash		: 1;	// Uncloaking, portalling, etc.
	unsigned UNKNOWN1		: 1;
	unsigned UFO			: 1;
	unsigned UNKNOWN2		: 1;

	PositionInfo()
	{
	}

	PositionInfo(PositionInfo const &Position)
	{
		memcpy(this, &Position, 1);
	}
};

PositionInfo MakePositionInfo(bool nStealth, bool nCloak, bool nXRadar, bool nAntiwarp,
							  bool nWarpFlash, bool nU1, bool nUFO, bool nU2)
{
	PositionInfo p;
	p.Stealth		= (nStealth ? 1 : 0);
	p.Cloaked		= (nCloak ? 1 : 0);
	p.XRadar		= (nXRadar ? 1 : 0);
	p.Antiwarp		= (nAntiwarp ? 1 : 0);
	p.WarpFlash		= (nWarpFlash ? 1 : 0);
	p.UNKNOWN1		= (nU1 ? 1 : 0);
	p.UFO			= (nUFO ? 1 : 0);
	p.UNKNOWN2		= (nU2 ? 1 : 0);
	return p;
}


// Macros

DWORD GetLong(BYTE * Message, const DWORD Offset)
{	// Get a LONG from a char *
	return *(LONG*)&(Message[Offset]);
}

WORD GetShort(BYTE * Message, const DWORD Offset)
{	// Get a SHORT from a char *
	return *(SHORT*)&(Message[Offset]);
}

DWORD GetLong(char * Message, const DWORD Offset)
{	// Get a LONG from a char *
	return *(LONG*)&(Message[Offset]);
}

WORD GetShort(char * Message, const DWORD Offset)
{	// Get a SHORT from a char *
	return *(SHORT*)&(Message[Offset]);
}

BYTE GetByte(char * Message, const DWORD Offset)
{	// Get a BYTE from a char *
	return *(BYTE*)&(Message[Offset]);
}

bool MatchStringCS(const char * lparam, const char * rparam, int length)
{	// Case-Sensitive string matching
	length--;

	for (SHORT index = 0; lparam[index] == rparam[index]; index++)
	{
		if ((lparam[index] == 0) || (index == length))
			return true;
	}

	return false;
}

bool MatchStringCI(const char * lparam, const char * rparam, int length)
{	// Case-Insensitive string matching
	BYTE LChar, RChar;
	length--;

	for (SHORT index = 0; index < length; index++)
	{
		LChar = lparam[index];

		if (LChar == 0)
			break;

		RChar = rparam[index];

		if ((LChar >= 65) && (LChar <= 90))
			LChar += 32;
		if ((RChar >= 65) && (RChar <= 90))
			RChar += 32;

		if (LChar != RChar)
			return false;
	}

	return true;
}

inline int TrimInt(int Value, int Limit)
{
	if (Value > Limit)
		return Limit;
	else
		return Value;
}

bool InvalidUsername(char * Username)
{
	if ( (Username[0] > 47 && Username[0] < 58) ||
		 (Username[0] > 64 && Username[0] < 91) ||
		 (Username[0] > 96 && Username[0] < 123) )		// The first character is alphanumeric
	{
		int i = 1;
		while (Username[i])
		{
			if (Username[i] < 32 || Username[i] > 127 || Username[i] == ':' || Username[i] == '%')	// Other characters are printable
				return true;

			i++;
		}
	}

	return false;
}


// Packet constructors

SHORT MakeSessionKey		(BYTE * Message, LONG Key)
{
	Message[0] = 0;
	Message[1] = 2;
	*(LONG*)&Message[2] = Key;

	return 6;
}

SHORT MakeEncryptRequest	(BYTE * Message, bool DisableEncryption)
{
	Message[0] = 0;
	Message[1] = 1;
	if (DisableEncryption)
		*(LONG*)&Message[2] = 0;
	else
		*(LONG*)&Message[2] = GetTickCount() ^ 0xAAAAAAAA;
	Message[6] = 1;
	Message[7] = 0;

	return 8;
}

SHORT MakeChatMessage		(BYTE * Message, BYTE ChatType, BYTE SoundCode,
							 SHORT TargetPlayer, const char * Text)
{
	int len = strlen(Text);

	Message[0] = 6;
	Message[1] = ChatType;
	Message[2] = SoundCode;
	*(SHORT*)&Message[3] = TargetPlayer;
	memcpy(&Message[5], Text, len + 1);

	return len + 6;
}

SHORT MakeRegFormResponse	(BYTE * Message, const char * RealName, const char * Email, const char * City,
							 const char * State, bool SexIsMale, BYTE Age, bool PlayAtHome, bool PlayAtWork,
							 bool PlayAtSchool, SHORT ProcessorType, DWORD Unk163, DWORD Unk167)
{
	/* Offset	Length	Description
	 * 0		1		Type byte
	 * 1		32		Real name
	 * 33		64		Email
	 * 97		32		City
	 * 129		24		State
	 * 153		1		Sex('M'/'F')
	 * 154		1		Age
	 * 155		1		Connecting from... home
	 * 156		1						   work
	 * 157		1						   school
	 * 158		2		Processor type (586)
	 * 160		3		?
	 * 163		4		? Looks pretty random...
	 * 167		4		? Maybe some extra ban ID
	 * 171		597		Hardware Information
	 * Windows NT-based OS's do not send any hardware information.
	 */

	memset(Message, 0, 768);

	Message[0] = 0x17;
	memcpy(&Message[1], RealName, strlen(RealName));
	memcpy(&Message[33], Email, strlen(Email));
	memcpy(&Message[97], City, strlen(City));
	memcpy(&Message[129], State, strlen(State));
	if (SexIsMale)
		Message[153] = 'M';
	else
		Message[153] = 'F';
	Message[154] = Age;
	Message[155] = (PlayAtHome ? 1 : 0);
	Message[156] = (PlayAtWork ? 1 : 0);
	Message[157] = (PlayAtSchool ? 1 : 0);
	*(SHORT*)&Message[158] = ProcessorType;
	*(LONG*)&Message[163] = Unk163;
	*(LONG*)&Message[167] = Unk167;

	return 768;
}

SHORT MakeACK				(BYTE * Message, LONG ID)
{
	Message[0] = 0;
	Message[1] = 4;
	*(LONG*)&Message[2] = ID;

	return 6;
}

SHORT MakeKoTHReset			(BYTE * Message)
{
	Message[0] = 30;

	return 1;
}

SHORT MakeSecurityChecksum	(BYTE * Message, LONG ParameterChecksum, LONG EXEChecksum, LONG MapChecksum,
							 SHORT LastPing, SHORT AveragePing, SHORT LowestPing, SHORT HighestPing,
							 DWORD WeaponCount, WORD S2CSlowCurrent, WORD S2CFastCurrent,
							 DWORD S2CSlowTotal, DWORD S2CFastTotal, bool SlowFrame)
{
// 00 01		  05		  09		  13		  17	19	  21	23	  25	27	  29	31	  33	35	  37	39
// 1A 19 02 00 00 1F D7 A8 42 FC 6B 10 63 3D 8D 1E BD 00 00 00 00 4B 04 00 00 00 00 4B 04 0E 00 CE CA 0C 00 0C 00 0D 00 00 
// 1A 5A 06 00 00 B9 A4 71 01 56 3E 6D 46 4D 9E 2D 31 00 00 00 00 A9 0B 00 00 00 00 5E 07 0E 00 CE CA 0C 00 0C 00 0F 00 00 
// 1A E1 03 00 00 0D FD 63 97 33 7F F7 70 ED 78 F7 0E 00 00 00 00 F3 03 00 00 00 00 F3 03 10 00 15 00 15 00 15 00 16 00 00 
// 1A C2 0B 00 00 52 27 63 8A 78 D9 51 18 82 04 69 34 00 00 00 00 D6 0B 00 00 00 00 E3 07 11 00 14 00 16 00 14 00 19 00 00 
	Message[0]				= 26;
	*(DWORD*)&Message[1]	= WeaponCount;
	*(DWORD*)&Message[5]	= ParameterChecksum;
	*(DWORD*)&Message[9]	= EXEChecksum;
	*(DWORD*)&Message[13]	= MapChecksum;
	*(DWORD*)&Message[17]	= S2CSlowTotal;
	*(DWORD*)&Message[21]	= S2CFastTotal;
	*(WORD*)&Message[25]	= S2CSlowCurrent;
	*(WORD*)&Message[27]	= S2CFastCurrent;
	*(WORD*)&Message[29]	= 14;				// ?
	*(WORD*)&Message[31]	= LastPing;
	*(WORD*)&Message[33]	= AveragePing;
	*(WORD*)&Message[35]	= LowestPing;
	*(WORD*)&Message[37]	= HighestPing;
	Message[39]				= (SlowFrame ? 1 : 0);

	return 40;
}

SHORT MakeLogin				(BYTE * Message, const char * Username, const char * Password,
							 LONG MachineID, SHORT TZB, LONG PermissionID, bool NewUser,
							 BYTE Options)
{
	memset(Message, 0, 101);

	Message[0]				= 9;
	Message[1]				= (NewUser ? 1 : 0);
	memcpy(&Message[2], Username, strlen(Username));
	memcpy(&Message[34], Password, strlen(Password));
	*(DWORD*)&Message[66]	= MachineID;
	Message[70]				= Options;
	*(SHORT*)&Message[71]	= TZB;
	*(SHORT*)&Message[73]	= 0x6f9d;
	*(SHORT*)&Message[75]	= 0x86;
	*(LONG*)&Message[77]	= 0x1bc;
	*(LONG*)&Message[81]	= 0x22b;
	*(DWORD*)&Message[85]	= PermissionID;

	return 101;
}

SHORT MakeBallRequest		(BYTE * Message, BYTE BallID, LONG ServerTime)
{
	Message[0] = 32;
	Message[1] = BallID;
	*(LONG*)&Message[2] = ServerTime;

	return 6;
}

SHORT MakeBallFire			(BYTE * Message, BYTE BallID, LONG ServerTime, SHORT X, SHORT Y)
{
	memset(Message, 0, 16);

	Message[0] = 31;
	Message[1] = BallID;
	*(SHORT*)&Message[2] = X;
	*(SHORT*)&Message[4] = Y;
	*(LONG*)&Message[12] = ServerTime;

	return 16;
}

SHORT MakeFlagRequest		(BYTE * Message, SHORT FlagID)
{
	Message[0] = 19;
	*(SHORT*)&Message[2] = FlagID;

	return 3;
}

SHORT MakeDropAllFlags		(BYTE * Message)
{
	Message[0] = 21;

	return 1;
}

SHORT MakeMapfileRequest	(BYTE * Message)
{
	Message[0] = 12;

	return 1;
}

SHORT MakeNewsfileRequest	(BYTE * Message)
{
	Message[0] = 13;

	return 1;
}

SHORT MakeArenaLogin		(BYTE * Message, ShipClasses ShipType, const char * ArenaName, WORD XRes, WORD YRes)
{
	Message[0] = 1;
	Message[1] = ShipType;
	Message[2] = 1;
	Message[3] = 0;
	*(WORD*)&Message[4] = XRes;
	*(WORD*)&Message[6] = YRes;
	if (strlen(ArenaName) == 0)
	{
		// Random main arena
		*(WORD*)&Message[8] = 0xFFFF;
		memset(&Message[10], 0, 16);
	}
	else
	{
		if (*ArenaName == '@')
		{
			// Specific main arena
			WORD ArenaNumber = ArenaName[1];

			ArenaNumber = ((ArenaNumber >= 48) || (ArenaNumber <= 57)) ? ArenaNumber - 48 : 0;

			*(WORD*)&Message[8] = ArenaNumber;
			memset(&Message[10], 0, 16);
		}
		else
		{
			// Private arena
			*(WORD*)&Message[8] = 0xFFFD;
			memcpy(&Message[10], ArenaName, 16);
		}
	}

	return 26;
}

SHORT MakeAttachRequest		(BYTE * Message, SHORT PlayerID)
{
	Message[0] = 16;
	*(SHORT*)&Message[1] = PlayerID;

	return 3;
}

SHORT MakeLeaveArena		(BYTE * Message)
{
	Message[0] = 2;

	return 1;
}


// Position packet checksum generator

BYTE SimpleChecksum(BYTE * Buffer, DWORD Length)
{
	BYTE Checksum = 0;

	for (DWORD i = 0; i < Length; i++)
		Checksum ^= Buffer[i];

	return Checksum;
}


SHORT MakePosition			(BYTE * Message, BYTE Direction, DWORD ServerTime, SHORT ThrustY,
							 SHORT Y, PositionInfo Toggle, SHORT X, SHORT ThrustX, SHORT Bounty,
							 SHORT Energy, WeaponInfo Weapon, SHORT S2CLag, SHORT Timer, ItemInfo II)
{
	memset(Message, 0, 32);

	// Required
	Message[0] = 3;
	Message[1] = Direction;
	*(LONG*)&Message[2] = ServerTime;
	*(SHORT*)&Message[6] = ThrustX;
	*(SHORT*)&Message[8] = Y;
	Message[10] = 0;
	memcpy(&Message[11], &Toggle, 1);
	*(SHORT*)&Message[12] = X;
	*(SHORT*)&Message[14] = ThrustY;
	*(SHORT*)&Message[16] = Bounty;
	*(SHORT*)&Message[18] = Energy;
	memcpy(&Message[20], &Weapon, 2);

	Message[10] = SimpleChecksum(Message, 22);

	// Extended
	*(SHORT*)&Message[22] = Energy;
	*(SHORT*)&Message[24] = S2CLag / 10;
	*(SHORT*)&Message[26] = Timer;
	memcpy(&Message[28], &II, 4);

	return 32;
}

SHORT MakeSyncResponse		(BYTE * Message, DWORD ClientTime, DWORD LocalTime)
{
	Message[0] = 0;
	Message[1] = 6;
	*(long*)&Message[2] = ClientTime;
	*(long*)&Message[6] = LocalTime;

	return 10;
}

SHORT MakeSyncRequest		(BYTE * Message, DWORD LocalTime, DWORD TotalSent, DWORD TotalRecvd)
{
	Message[0] = 0;
	Message[1] = 5;
	*(long*)&Message[2] = LocalTime;
	*(long*)&Message[6] = TotalSent;
	*(long*)&Message[10] = TotalRecvd;

	return 14;
}


// SubSpace Encryption Table

class SSET
{
public:
	DWORD SSKey;
	BYTE SSBuf[520];

	SSET()
	{
		SSKey = 0;
	}
	inline SSET(const SSET &copy)
	{
		SSKey = copy.SSKey;
		memcpy(SSET::SSBuf, copy.SSBuf, 520);
	}
};


// Utilities

LONG ResolveHostname(char * Hostname)
{
	LONG RetIP = inet_addr(Hostname);

	if (RetIP == -1)
	{
		hostent * host = gethostbyname(Hostname);

		if (host)
			RetIP = *((LONG*)*(host->h_addr_list));
		else
			RetIP = -1;
	}

	if (RetIP == -1)
	{
		_string S;
		S = "Unable to resolve hostname: ";
		S += Hostname;
		MessageBox(NULL, S.Text, "ResolveHostname():", 0);
	}

	return RetIP;
}


// Packet contents

void DumpHex(BYTE * Offset, LONG Length)
{
/*	printf("0000  ");
	for(int cnt = 0; cnt < Length; cnt++)
	{
		if(cnt && cnt % 16 == 0);
			printf("\n%04X  ", cnt);
		printf("%02X ", Offset[cnt]);
	}
	printf("\n");		*/
}

class PacketStruct
{
public:
	/* Contents */
	BotSession      * Source;	// Source of packet
	DWORD			  Length;	// Length of the packet
	BYTE			* Message;	// Contents of the packet (unencrypted)

	/* Parsers */
	PacketStruct(BotSession * NewSource, BYTE * NewMessage, DWORD NewLength)
	{
		Message = new BYTE[NewLength];
		memcpy(Message, NewMessage, NewLength);
		Length = NewLength;
		Source = NewSource;
	}

	~PacketStruct()
	{
		delete Message;
	}

	inline PacketStruct(const PacketStruct &copy)
	{
		Source = copy.Source;
		Length = copy.Length;
		Message = new BYTE[Length];
		memcpy(Message, copy.Message, Length);
	}

	void Dump()
	{
		DumpHex(Message, Length);
	}
};


// Packet contents

class PostlogNode
{
public:
	/* Contents */
	DWORD	Length;		// Length of the packet
	BYTE  * Message;	// Contents of the packet (encrypted)
	DWORD	PacketID;	// ID of packet to be ACK'd
	LONG	FirstSend;	// 'Tick count from first send
	LONG	LastSend;	// 'Tick count from last send

	/* Parsers */
	void SetTime()
	{
		LastSend = GetTickCount();
	}

	PostlogNode(BYTE * NewMessage, DWORD NewLength, LONG NewPacketID)
	{	// NewMessage encrypted PRIOR to call
		Length		= NewLength;
		PacketID	= NewPacketID;
		LastSend	= GetTickCount();
		FirstSend	= LastSend;
		Message		= new BYTE[NewLength];
		memcpy(Message, NewMessage, Length);
	}

	~PostlogNode()
	{
		delete Message;
	}

	PostlogNode(const PostlogNode &copy)
	{
		Length = copy.Length;
		Message = new BYTE[Length];
		memcpy(Message, copy.Message, Length);
		PacketID = copy.PacketID;
		LastSend = copy.LastSend;
	}
};


// Chunks! (Recving)

class IncomingChunk
{
public:
	char  * Contents;
	DWORD	Length;

	IncomingChunk()
	{
		Contents = NULL;
		Length = 0;
	}

	~IncomingChunk()
	{
		if (Contents)
			Clear();
	}

	void Set(char * Message, DWORD mLen)
	{
		Length = mLen;
		Contents = new char[mLen];
		memcpy(Contents, Message, mLen);
	}

	void Append(char * Message, DWORD mLen)
	{
		if (Contents)
		{
			char * OldContents = Contents;
			LONG NewLength = Length + mLen;
			Contents = new char[NewLength];
			memcpy(Contents, OldContents, Length);
			memcpy(&Contents[Length], Message, mLen);
			Length = NewLength;
		}
		else
		{
			Set(Message, mLen);
		}
	}

	void Clear()
	{
		delete Contents;
		Contents = NULL;
		Length = 0;
	}
};


// Ship settings

class ShipSettings						// 144 bytes wide
{
public:
	SHORT SuperTime;
	SHORT UNKNOWN0;
	SHORT ShieldsTime;
	SHORT UNKNOWN1;
	SHORT Gravity;
	SHORT GravityTopSpeed;
	SHORT BulletFireEnergy;
	SHORT MultiFireEnergy;
	SHORT BombFireEnergy;
	SHORT BombFireEnergyUpgrade;
	SHORT LandmineFireEnergy;
	SHORT LandmineFireEnergyUpgrade;
	SHORT BulletSpeed;
	SHORT BombSpeed;
	SHORT UNKNOWN2;
	SHORT MultiFireAngle;
	SHORT CloakEnergy;
	SHORT StealthEnergy;
	SHORT AntiWarpEnergy;
	SHORT XRadarEnergy;
	SHORT MaximumRotation;
	SHORT MaximumThrust;
	SHORT MaximumSpeed;
	SHORT MaximumRecharge;
	SHORT MaximumEnergy;
	SHORT InitialRotation;
	SHORT InitialThrust;
	SHORT InitialSpeed;
	SHORT InitialRecharge;
	SHORT InitialEnergy;
	SHORT UpgradeRotation;
	SHORT UpgradeThrust;
	SHORT UpgradeSpeed;
	SHORT UpgradeRecharge;
	SHORT UpgradeEnergy;
	SHORT AfterburnerEnergy;
	SHORT BombThrust;
	SHORT BurstSpeed;
	SHORT TurretThrustPenalty;
	SHORT TurretSpeedPenalty;
	SHORT BulletFireDelay;
	SHORT MultiFireDelay;
	SHORT BombFireDelay;
	SHORT LandmineFireDelay;
	SHORT RocketTime;
	SHORT InitialBounty;
	SHORT DamageFactor;
	SHORT PrizeShareLimit;
	SHORT AttachBounty;
	SHORT SoccerThrowTime;
	SHORT SoccerBallFriction;
	SHORT SoccerBallProximity;
	SHORT SoccerBallSpeed;

	BYTE TurretLimit;
	BYTE BurstShrapnel;
	BYTE MaxMines;
	BYTE RepelMax;
	BYTE BurstMax;
	BYTE DecoyMax;
	BYTE ThorMax;
	BYTE BrickMax;
	BYTE RocketMax;
	BYTE PortalMax;
	BYTE InitialRepel;
	BYTE InitialBurst;
	BYTE InitialBrick;
	BYTE InitialRocket;
	BYTE InitialThor;
	BYTE InitialDecoy;
	BYTE InitialPortal;
	BYTE BombBounceCount;

	unsigned ShrapnelMax	: 4;
	unsigned ShrapnelRate	: 4;

	unsigned SuperStatus	: 2;
	unsigned CloakStatus	: 2;
	unsigned StealthStatus	: 2;
	unsigned XRadarStatus	: 2;

	unsigned AntiwarpStatus	: 2;
	unsigned InitialGuns	: 2;
	unsigned MaxGuns		: 2;
	unsigned InitialBombs	: 2;

	unsigned MaxBombs		: 2;
	unsigned DoubleBarrel	: 1;
	unsigned EmpBomb		: 1;
	unsigned SeeMines		: 1;
	unsigned UNKNOWN3		: 3;

	BYTE UNKNOWN4[16];
};


// Prize settings

class PrizeSettings						// 28 bytes wide
{
public:
	BYTE QuickCharge;
	BYTE Energy;
	BYTE Rotation;
	BYTE Stealth;
	BYTE Cloak;
	BYTE XRadar;
	BYTE Warp;
	BYTE Gun;
	BYTE Bomb;
	BYTE BouncingBullets;
	BYTE Thruster;
	BYTE TopSpeed;
	BYTE Recharge;
	BYTE Glue;
	BYTE MultiFire;
	BYTE Proximity;
	BYTE AllWeapons;
	BYTE Shields;
	BYTE Shrapnel;
	BYTE AntiWarp;
	BYTE Repel;
	BYTE Burst;
	BYTE Decoy;
	BYTE Thor;
	BYTE MultiPrize;
	BYTE Brick;
	BYTE Rocket;
	BYTE Portal;
};


// Arena settings

class ArenaSettings					// 1428 bytes wide
{
public:
	LONG ident;						// 0000 Constantly 0F 00 00 00
	ShipSettings ships[8];			// 0004
	LONG BulletDamageLevel;			// 1156 BulletDamageLevel * 1000
	LONG BombDamageLevel;			// 1160 BombDamageLevel * 1000
	LONG BulletAliveTime;			// 1164
	LONG BombAliveTime;				// 1168
	LONG DecoyAliveTime;			// 1172
	LONG SafetyLimit;				// 1176
	LONG FrequencyShift;			// 1180
	LONG MaxFrequency;				// 1184
	LONG RepelSpeed;				// 1188
	LONG MineAliveTime;				// 1192
	LONG BurstDamageLevel;			// 1196 BurstDamageLevel * 1000
	LONG BulletDamageUpgrade;		// 1200 BulletDamageUpgrade * 1000
	LONG FlagDropDelay;				// 1204
	LONG EnterGameFlaggingDelay;	// 1208
	LONG RocketThrust;				// 1212
	LONG RocketSpeed;				// 1216
	LONG InactiveShrapnelDamage;	// 1220
	LONG WormholeSwitchTime;		// 1224
	LONG ActivateAppShutdownTime;	// 1228
	LONG ShrapnelSpeed;				// 1232
	BYTE UNKNOWN0[16];				// 1236
	SHORT SendRoutePercent;			// 1252
	SHORT BombExplodeDelay;			// 1254
	SHORT SendPositionDelay;		// 1256
	SHORT BombExplodePixels;		// 1258
	SHORT DeathPrizeTime;			// 1260
	SHORT JitterTime;				// 1262
	SHORT EnterDelay;				// 1264
	SHORT EngineShutdownTime;		// 1266
	SHORT ProximityDistance;		// 1268
	SHORT BountyIncreaseForKill;	// 1270
	SHORT BounceFactor;				// 1272
	SHORT MapZoomFactor;			// 1274
	SHORT MaxBonus;					// 1276
	SHORT MaxPenalty;				// 1278
	SHORT RewardBase;				// 1280
	SHORT RepelTime;				// 1282
	SHORT RepelDistance;			// 1284
	SHORT UNKNOWN2;					// 1286
	SHORT FlaggerOnRadar;			// 1288
	SHORT FlaggerKillMultiplier;	// 1290
	SHORT PrizeFactor;				// 1292
	SHORT PrizeDelay;				// 1294
	SHORT PrizeMinimumVirtual;		// 1296
	SHORT PrizeUpgradeVirtual;		// 1298
	SHORT PrizeMaxExist;			// 1300
	SHORT PrizeMinExist;			// 1302
	SHORT PrizeNegativeFactor;		// 1304
	SHORT DoorDelay;				// 1306
	SHORT AntiwarpPixels;			// 1308
	SHORT DoorMode;					// 1310
	SHORT FlagBlankDelay;			// 1312
	SHORT NoDataFlagDropDelay;		// 1314
	SHORT MultiPrizeCount;			// 1316
	SHORT BrickTime;				// 1318
	SHORT WarpRadiusLimit;			// 1320
	SHORT EBombShutdownTime;		// 1322
	SHORT EBombDamagePercent;		// 1324
	SHORT RadarNeutralSize;			// 1326
	SHORT WarpPointDelay;			// 1328
	SHORT NearDeathLevel;			// 1330
	SHORT BBombDamagePercent;		// 1332
	SHORT ShrapnelDamagePercent;	// 1334
	SHORT ClientSlowPacketTime;		// 1336
	SHORT FlagDropResetReward;		// 1338
	SHORT FlaggerFireCostPercent;	// 1340
	SHORT FlaggerDamagePercent;		// 1342
	SHORT FlaggerBombFireDelay;		// 1344
	SHORT SoccerPassDelay;			// 1346
	SHORT SoccerBallBlankDelay;		// 1348
	SHORT S2CNoDataKickoutDelay;	// 1350
	SHORT FlaggerThrustAdjustment;	// 1352
	SHORT FlaggerSpeedAdjustment;	// 1354
	SHORT ClientSlowPacketSampleSize;//1356
	BYTE UNKNOWN3[10];				// 1358
	BYTE RandomShrapnel;			// 1368
	BYTE SoccerBallBounce;			// 1369
	BYTE SoccerAllowBombs;			// 1370
	BYTE SoccerAllowGuns;			// 1371
	BYTE SoccerMode;				// 1372
	BYTE MaxPerTeam;				// 1373
	BYTE MaxPerPrivateTeam;			// 1374
	BYTE TeamMaxMines;				// 1375
	BYTE WormholeGravityBombs;		// 1376
	BYTE BombSafety;				// 1377
	BYTE MessageReliable;			// 1378
	BYTE TakePrizeReliable;			// 1379
	BYTE AllowAudioMessages;		// 1380
	BYTE PrizeHideCount;			// 1381
	BYTE ExtraPositionData;			// 1382
	BYTE SlowFrameCheck;			// 1383
	BYTE CarryFlags;				// 1384
	BYTE AllowSavedShip;			// 1385
	BYTE RadarMode;					// 1386
	BYTE VictoryMusic;				// 1387
	BYTE FlaggerGunUpgrade;			// 1388
	BYTE FlaggerBombUpgrade;		// 1389
	BYTE SoccerUserFlagger;			// 1390
	BYTE SoccerBallLocation;		// 1391
	BYTE UNKNOWN4[8];				// 1392
	PrizeSettings pw;				// 1400
};


// Animating object descriptor

class Animation
{
public:
	long ra;
	DWORD LastUpdateTime;
	WORD AnimationFrame;
	WORD FrameRate;

	Animation()
	{
		AnimationFrame = 0;
		LastUpdateTime = GetTickCount();
		ra = 0;
		FrameRate = 100;
	}

	void SetFrameRate(WORD nFrameRate)
	{
		FrameRate = nFrameRate;
	}

	void AdvanceFrame()
	{
		long NewTime = GetTickCount();
		long DiffTime = NewTime - LastUpdateTime;
		LastUpdateTime = NewTime;

		AnimationFrame += (DiffTime + ra) / FrameRate;
		ra = (DiffTime + ra) % FrameRate;
	}
};


// Projectile object descriptor

class Projectile
{
	long LastX, LastY;

	bool CheckBounceTile(long &x1, long &y1, long &x2, long &y2, long x, long y, long TileX, long TileY)
	{
		TileX &= 1023;
		TileY &= 1023;

		// Check map tile type
		BYTE v = Map[(TileY << 10) + TileX];

		if ((v > 0) && (v < 170))	// All tiles up to the flag # 170
		{
			// Reflect thrust and projection over line of collision
			if (TileY != LastY)
			{
				// Struck the top or bottom of a tile
				ty = -ty;
				ry = -ry;
				y1 += SGN(y1 - y2);
				y2 += y1 - y2;
			}

			if (TileX != LastX)
			{
				// Struck the left or right of a tile
				tx = -tx;
				rx = -rx;
				x1 += SGN(x1 - x2);
				x2 += x1 - x2;
			}

			// Monitor number of times it has bounced
			if (BounceCounter > 0)
				BounceCounter--;

			BounceLimit--;

			// Struck a barrier
			return true;
		}

		return false;
	}

	/* Bounce Counter
	 * -2     = Goes thru walls
	 * -1     = No bounce limit
	 * 0      = Dead
	 * n	  = Bounces n - 1 times
	 */

	bool CheckCollision(long &x, long &y, long &x2, long &y2)
	{
		// Collision detection
		if (BounceCounter != -2)
		{
			register long ThisX = x >> 4;
			register long ThisY = y >> 4;

			if ((ThisX != LastX) || (ThisY != LastY))
			{
				if (CheckBounceTile(x, y, x2, y2, x, y, ThisX, ThisY))
					return true;

				LastX = ThisX;
				LastY = ThisY;
			}

			if (BounceProximity)
			{
				register long ThisXPlus = (x + BounceProximity) >> 4;
				register long ThisYPlus = (y + BounceProximity) >> 4;
				register long ThisXMinus = (x - BounceProximity) >> 4;
				register long ThisYMinus = (y - BounceProximity) >> 4;

				if (CheckBounceTile(x, y, x2, y2, x + BounceProximity, y, ThisXPlus, ThisY))
					return true;
				if (CheckBounceTile(x, y, x2, y2, x - BounceProximity, y, ThisXMinus, ThisY))
					return true;
				if (CheckBounceTile(x, y, x2, y2, x, y + BounceProximity, ThisX, ThisYPlus))
					return true;
				if (CheckBounceTile(x, y, x2, y2, x, y - BounceProximity, ThisX, ThisYMinus))
					return true;
//				if (CheckBounceTile(x, y, x2, y2, x + BounceProximity, y + BounceProximity, ThisXPlus, ThisYPlus))
//					return true;
//				if (CheckBounceTile(x, y, x2, y2, x - BounceProximity, y + BounceProximity, ThisXMinus, ThisYPlus))
//					return true;
//				if (CheckBounceTile(x, y, x2, y2, x + BounceProximity, y - BounceProximity, ThisXPlus, ThisYMinus))
//					return true;
//				if (CheckBounceTile(x, y, x2, y2, x - BounceProximity, y - BounceProximity, ThisXMinus, ThisYMinus))
//					return true;
			}
		}

		// Didn't hit anything
		return false;
	}

	bool TracePath(long &x, long &y, long &x2, long &y2)
	{
		if (BounceLimit == 0)
			return false;

		long dx = DISTANCE(x, x2),
			 dy = DISTANCE(y, y2);

		long acc = 0;

		long mx = SGN(x2 - x),
			 my = SGN(y2 - y);

		if (dx > dy)
		{
			while (x != x2)
			{
				x += mx;
				acc += dy;

				if (acc >= dx)
				{
					y += my;
					acc -= dx;
				}

				if (CheckCollision(x, y, x2, y2))
				{
					TracePath(x, y, x2, y2);
					return true;
				}
			}
		}
		else
		{
			while (y != y2)
			{
				y += my;
				acc += dx;

				if (acc >= dy)
				{
					x += mx;
					acc -= dy;
				}

				if (CheckCollision(x, y, x2, y2))
				{
					TracePath(x, y, x2, y2);
					return true;
				}
			}
		}

		return false;
	}

public:
	long tx, ty;
	long x, y;
	long rx, ry;
	long BounceCounter;		// 0 = Doesn't bounce, -1 = Bounces forever, -2 = Goes thru walls
	long BounceLimit;
	DWORD LastUpdateTime;
	long Friction, ah;
	long WallFriction;
	long BounceProximity;
	BYTE * Map;

	bool ValidateCurrentPosition()
	{
		if (GetTickCount() - LastUpdateTime > 3000)
			return false;
		else
			return true;
	}

	void SetBounce(long Bounce)
	{
		BounceCounter = Bounce;
	}

	void SetFriction(long nFriction)
	{
		Friction = nFriction;
	}

	void SetWallFriction(long nFriction)
	{
		WallFriction = nFriction;
	}

	void SetBounceProximity(long nProx)
	{
		BounceProximity = nProx;
	}

	Projectile()
	{
		x = y = 0;
		tx = ty = 0;
		rx = ry = 0;
		BounceCounter = 0;
		LastUpdateTime = 0;
		Friction = ah = 0;
		WallFriction = 16;
		BounceProximity = 0;
	}

	void UpdatePosition(DWORD Time, long nx, long ny, long ntx, long nty)
	{
		x = nx;
		y = ny;

		tx = ntx;
		ty = nty;

		rx = 0;
		ry = 0;

		LastUpdateTime = Time;
	}

	bool TraceProjectile(BYTE * nMap)
	{
		// Find time difference
		DWORD NewTime = GetTickCount();
		DWORD DiffTime = NewTime - LastUpdateTime;
		LastUpdateTime = NewTime;

		if (Friction)
		{
			// Calculate friction influence
			long fx = DiffTime * Friction + ah;
			long ux = fx / 30;
			ah = fx % 30;

			if (ux)
			{
				// Adjust thrust
				if (tx < 0)
				{
					tx += ux;
					if (tx > 0) tx = 0;
				}
				else
				{
					tx -= ux;
					if (tx < 0) tx = 0;
				}
				if (ty < 0)
				{
					ty += ux;
					if (ty > 0) ty = 0;
				}
				else
				{
					ty -= ux;
					if (ty < 0) ty = 0;
				}
			}
		}

		// Calculate projectile displacement
		long wx = DiffTime * tx + rx;
		long ThrustX = wx / 10000;
		rx = wx % 10000;

		long wy = DiffTime * ty + ry;
		long ThrustY = wy / 10000;
		ry = wy % 10000;

		// Generate a pair of target coords
		long x2 = x + ThrustX;
		long y2 = y + ThrustY;

		Map = nMap;

		// Trace a path to destination
		BounceLimit = 10;

		LastX = x >> 4;
		LastY = y >> 4;

		return TracePath(x, y, x2, y2);
	}
};


// Player

class Player : public Projectile
{
public:
	char Name[20];
	char Squadname[20];
	BYTE Banner[96];

	BYTE ShipType;		// Current ship type

	long cEnergy;		// Current energy
	long cRotation;		// Current rotation
	long cThrust;		// Current acceleration
	long cSpeed;		// Current velocity
	long cRecharge;		// Current rate of recharge
	long cBounty;		// Current bounty
	long cMines;		// Total mines

	long mEnergy;		// Max energy
	long mThrust;		// Max thrust
	long mSpeed;		// Max speed
	long mRecharge;		// Max recharge

	DWORD tBomb;		// Last bomb time
	DWORD tBullet;		// Last bullet time
	DWORD tRocket;		// Last rocket time
	DWORD tKing;		// Last KoTH reset
	DWORD tFlag;		// Last flag timer reset
	DWORD tSoccer;		// Last soccer ball claim
	DWORD tSuper;		// Last super charge
	DWORD tShields;		// Last shields charge

	BYTE cRepel;		// Number of repels in stock
	BYTE cBurst;		// Number of bursts available
	BYTE cDecoy;		// Decoy count
	BYTE cThor;			// Total number of thors
	BYTE cBrick;		// Bricks in storage
	BYTE cRocket;		// Rockets available
	BYTE cPortal;		// Portal count
	BYTE cShrap;		// Amount of shrap produced by bombs

	BYTE lGun;			// Gun level
	BYTE lBomb;			// Bomb level

	bool bSuper;		// *super on/off
	bool bUFO;			// *ufo on/off
	bool bCloak;		// Cloak on/off
	bool bStealth;		// Stealth on/off
	bool bXRadar;		// X-Radar on/off
	bool bAntiwarp;		// Antiwarp on/off
	bool bMultifire;	// Multifire on/off

	bool hSuper;		// Has *super? - set by server in settings
	bool hUFO;			// Has *ufo? - set by server in a packet
	bool hCloak;		// Has Cloak?
	bool hStealth;		// Has Stealth?
	bool hXRadar;		// Has X-Radar?
	bool hAntiwarp;		// Has Antiwarp?
	bool hMultifire;	// Has Multifire?
	bool hProxbombs;	// Proximity bombs on/off?
	bool hBouncingBullets;// Bouncing bullets on/off?
	bool hDoubleBarrel;	// Double barrel
	bool hEMP;			// EMP bomb
	bool hSeeMines;		// Ability to see mines on radar

	ShipSettings * Settings;

	void SetSettings(ShipSettings * nSettings)
	{
		Settings = nSettings;
	}

	void ResetShip()
	{
		if (Settings)
		{
			cRotation	= Settings->InitialRotation;
			cThrust		= Settings->InitialThrust;
			cSpeed		= Settings->InitialSpeed;
			cRecharge	= Settings->InitialRecharge;
			cEnergy		= Settings->InitialEnergy;
			cBounty		= Settings->InitialBounty;
			cMines		= 0;

			mEnergy		= cEnergy;
			mThrust		= cThrust;
			mSpeed		= cSpeed;
			mRecharge	= cRecharge;

			tBomb = tBullet = tRocket = tKing = tFlag = tSoccer = tSuper = tShields = 0;

			cRepel	= Settings->InitialRepel;
			cBurst	= Settings->InitialBurst;
			cBrick	= Settings->InitialBrick;
			cRocket	= Settings->InitialRocket;
			cThor	= Settings->InitialThor;
			cDecoy	= Settings->InitialDecoy;
			cPortal	= Settings->InitialPortal;
			cShrap	= 0;

			bSuper		= false;
			bUFO		= false;
			bCloak		= false;
			bStealth	= false;
			bXRadar		= false;
			bAntiwarp	= false;
			bMultifire	= false;

			hSuper			= (Settings->SuperStatus == 2 ? true : false);
			hUFO			= false;
			hCloak			= (Settings->CloakStatus == 2 ? true : false);
			hStealth		= (Settings->StealthStatus == 2 ? true : false);
			hXRadar			= (Settings->XRadarStatus == 2 ? true : false);
			hAntiwarp		= (Settings->AntiwarpStatus == 2 ? true : false);
			hMultifire		= false;
			hProxbombs		= false;
			hBouncingBullets= false;
			hDoubleBarrel	= Settings->DoubleBarrel;
			hEMP			= Settings->EmpBomb;
			hSeeMines		= Settings->SeeMines;

			lGun	= Settings->InitialGuns;
			lBomb	= Settings->InitialBombs;
		}
	}

	LONG	 FlagPoints;	// Points resulting from flag victories
	LONG	 KillPoints;	// Points resulting from kills
	SHORT	 Frequency;		// Frequency
	SHORT	 Wins;			// Kills
	SHORT	 Losses;		// Deaths
	BYTE	 D;				// Direction
	BYTE	 Ping;			// Server-provided player latency
	BYTE	 PlayerID;		// PlayerID
	Player * AttachedTo;	// Player attached to

	Player() : Projectile()
	{
		memset(Name,		0, 20);
		memset(Squadname,	0, 20);
		memset(Banner,		0, 96);

		D			= 0;
		Ping		= 0;
		AttachedTo	= NULL;
	}
};


// Power ball descriptor

class Powerball : public Projectile, public Animation
{
public:
	Player * Controller;
	SHORT	 BallID;
	bool	 Carried;

	LONG	 LastSend;
	DWORD	 TimeStamp;
};


// Flag descriptor

class FlagData : public Animation
{
public:
	Player *Controller;
	SHORT	FlagID;
	long	x;
	long	y;
	SHORT	Team;

	LONG	LastSend;
	bool	Carried;
};


// Brick descriptor

class Brick : public Animation
{
public:
	long	x;
	long	y;
	SHORT	Team;
	DWORD	DestroyTime;
};


// Weapon descriptor

class Weapon : public Projectile, public Animation
{
public:
	WeaponInfo Type;

	Player * Owner;
	DWORD DestroyTime;
};


// Outgoing clustered packet

struct LoggedPacket
{
	BYTE * Message;
	DWORD Length;
	bool Reliable;

	LoggedPacket(BYTE * msg, DWORD len, bool rel)
	{
		Message = new BYTE[len];
		memcpy(Message, msg, len);
		Length = len;
		Reliable = rel;
	}

	~LoggedPacket()
	{
		delete Message;
	}
};


// Weapons impact effects

class WeaponImpact : public Animation
{
public:
	WeaponImpact(long nx, long ny, BYTE nWeaponType, DWORD nTotalFrames)
	{
		TotalFrames = nTotalFrames;
		x = nx;
		y = ny;
		WeaponType = nWeaponType;
	}

	long x;
	long y;
	BYTE WeaponType;
	DWORD TotalFrames;
};


// Bot object

#include "settings.h"

class BotSession : public Mutex, public Animation
{
	//////// GFX ////////
	DXWindow Window;

	_font Font;

	_chat Chat;

	_starfield Starfield;

	_tileset TileSet;
	_tileset ShipSet;
	_tileset Asteroids;
	_tileset Asteroids2;
	_tileset Station;
	_tileset Wormhole;
	_tileset BombSet;
	_tileset BulletSet;
	_tileset FlagSet;
	_tileset MineSet;
	_tileset BrickSet;
	_tileset ShrapSet;
	_tileset BallSet;
	_tileset GoalSet;
	_tileset RepelSet;
	_tileset GreenSet;
	_tileset TurretSet;
	_tileset ExplosionSet;
	_tileset ImpactSet;

	_tileset StatusBox;
	_tileset NumberFont1;
	_tileset NumberFont2;
	_numbers NumberSet1;
	_numbers NumberSet2;

	_tileset ItemSet;
	_tileset ItemFont;

	ChatEntryBox ChatBox;

	_radar Radar;

	DWORD MapWidth, MapHeight;		// Map dimensions in pixels

	void DrawTiledBackdrop();
	inline void DrawTile(DWORD x, DWORD y, DWORD DestX, DWORD DestY, DWORD SrcX, DWORD SrcY, DWORD W, DWORD H);
	inline void DrawTile(DWORD x, DWORD y, DWORD DestX, DWORD DestY);

	//////// Threads ////////
	DWORD ThreadsOpen;

	HANDLE hResend,		// Handle to the resend thread
		   hReceive,	// Handle to the receive thread
		   hGraphics;	// Handle to the graphics thread

	void ResendThread();
	void RecvThread();
	void GraphicsThread();

	friend DWORD WINAPI CallResendThread(BotSession * ThisInstance)
	{
		ThisInstance->ThreadsOpen++;

			ThisInstance->ResendThread();

		ThisInstance->ThreadsOpen--;

		return 0;
	}

	friend DWORD WINAPI CallRecvThread(BotSession * ThisInstance)
	{
		ThisInstance->ThreadsOpen++;

			ThisInstance->RecvThread();

		ThisInstance->ThreadsOpen--;

		return 0;
	}

	friend DWORD WINAPI CallGraphicsThread(BotSession * ThisInstance)
	{
		ThisInstance->ThreadsOpen++;

			ThisInstance->GraphicsThread();

		ThisInstance->ThreadsOpen--;

		return 0;
	}

	//////// Handlers ////////

	// Unsure about these
	friend bool HandleMapRequest		(PacketStruct * Packet);	// Giving the go-ahead for map transmission
	friend bool HandleLoginUnknown		(PacketStruct * Packet);	// 1 byte long, sent with password response
	friend bool HandleStatusUnknown		(PacketStruct * Packet);	// 5 bytes long, at least one field is a PlayerID

	// Game-specific protocol
	friend bool HandleUnknown			(PacketStruct * Packet);	// Unknown message type (debug output)
	friend bool HandleSpecial			(PacketStruct * Packet);	// Privelaged types
	friend bool HandlePasswordAccept	(PacketStruct * Packet);	// Password packet response
	friend bool HandleIDChange			(PacketStruct * Packet);	// Player ID change
	friend bool HandleArenaSettings		(PacketStruct * Packet);	// Arena settings
	friend bool HandlePlayerEntering	(PacketStruct * Packet);	// Player entering arena
	friend bool HandleInGameFlag		(PacketStruct * Packet);	// Position packets may begin
	friend bool HandleDoorToggle		(PacketStruct * Packet);	// Toggle doors
	friend bool HandleMapInformation	(PacketStruct * Packet);	// Map checksum
	friend bool HandleFlagPosition		(PacketStruct * Packet);	// Flag coordinates & carriers
	friend bool HandleKeepAlive			(PacketStruct * Packet);	// Ping? Pong!
	friend bool HandlePlayerLeaving		(PacketStruct * Packet);	// Delete player by ID
	friend bool HandlePlayerBanner		(PacketStruct * Packet);	// Change player banner
	friend bool HandleMapFile			(PacketStruct * Packet);	// Got map file!
	friend bool HandleSecurityChecksum	(PacketStruct * Packet);	// Security checksum request
	friend bool HandleSetFreq			(PacketStruct * Packet);	// Set player frequency
	friend bool HandleSetFreqAndShip	(PacketStruct * Packet);	// Set player ship type and frequency
	friend bool HandleSpecPlayer		(PacketStruct * Packet);	// Set player ship type to 8
	friend bool HandleWeaponsFire		(PacketStruct * Packet);	// Someone shot a projectile
	friend bool HandlePlayerDeath		(PacketStruct * Packet);	// Increment player kills & deaths
	friend bool HandleChat				(PacketStruct * Packet);	// Subclassed within the message body
	friend bool HandleScoreUpdate		(PacketStruct * Packet);	// Modify player scores
	friend bool HandleSoccerGoal		(PacketStruct * Packet);	// Goal! Goal! Goal! Goal!
	friend bool HandleKoTHReset			(PacketStruct * Packet);	// KoTH reset
	friend bool HandlePlayerPosition	(PacketStruct * Packet);	// A player has changed coordinates
	friend bool HandleBallPosition		(PacketStruct * Packet);	// Global ball positioning system
	friend bool HandleArenaList			(PacketStruct * Packet);	// Arena response packet
	friend bool HandleFlagReward		(PacketStruct * Packet);	// Grant point flag reward points
	friend bool HandleFlagVictory		(PacketStruct * Packet);	// Grant point flag victory points
	friend bool HandleFlagClaim			(PacketStruct * Packet);	// Grab a flag
	friend bool HandleFlagDrop			(PacketStruct * Packet);	// Drop a flag
	friend bool HandleScoreReset		(PacketStruct * Packet);	// Delete player score(s)
	friend bool HandleToggleUFO			(PacketStruct * Packet);	// Toggle *ufo status
	friend bool HandleTurretPlayer		(PacketStruct * Packet);	// Set player turret
	friend bool HandleShipReset			(PacketStruct * Packet);	// Reset ship to settings
	friend bool HandlePrize				(PacketStruct * Packet);	// React to a prize
	friend bool HandleTurfFlagStatus	(PacketStruct * Packet);	// Update flags in a turf-style game
	friend bool HandleFileTransfer		(PacketStruct * Packet);	// File transfer
	friend bool HandlePrizePickup		(PacketStruct * Packet);	// Ping data
	friend bool HandleUnturret			(PacketStruct * Packet);	// Remove player turret
	friend bool HandleBannerAds			(PacketStruct * Packet);	// "SSCE Chaos Zone SVS" zone's banners
	friend bool HandleCustomMessage		(PacketStruct * Packet);	// Sent by SSC biller for busy signal

	// SubSpace core protocol
	friend bool HandleUnknownSpecial	(PacketStruct * Packet);	// Unhandled packet
	friend bool HandleEncryptRequest	(PacketStruct * Packet);	// Request to encrypt communication
	friend bool HandleEncryptResponse	(PacketStruct * Packet);	// Session key
	friend bool HandleReliable			(PacketStruct * Packet);	// Message with a reliable header
	friend bool HandleACK				(PacketStruct * Packet);	// ACK message
	friend bool HandleSyncRequest		(PacketStruct * Packet);	// Synchronization request
	friend bool HandleSyncResponse		(PacketStruct * Packet);	// Synchronization response
	friend bool HandleDisconnect		(PacketStruct * Packet);	// Bye!
	friend bool HandleCluster			(PacketStruct * Packet);	// Cluster packet
	friend bool HandleSmallChunkBody	(PacketStruct * Packet);	// Host is blowing chunks
	friend bool HandleSmallChunkTail	(PacketStruct * Packet);	// More chunks
	friend bool HandleBigChunk			(PacketStruct * Packet);	// Host is blowing really big chunks

	//////// Identifiers ////////
	BYTE	MyID;			// Bot player ID
	DWORD	TimeStamp;		// Used to find 'ticks since the last recv'd packet
	DWORD	SyncTime;		// If this is non-zero, SyncResponses will be checked for validity
	long	MachineID;		// Volume C:\ serial number
	long	PermissionID;	// Random number 'D1' in registry
	SHORT	TZB;			// TimeZone bias
	char	BotName[32];	// Bot name
	char	BotPassword[32];// Bot password

	//////// Connection status ////////
	bool	PasswordWait;	// Currently waiting for a password to be validated?
	bool	SendingPosition;// May I send any position packets?
	bool	HasSession;		// 00 01 / 00 02 initiation protocol completed?
	DWORD	LastSessionSend;// Last time a 00 01 packet was sent.

	//////// Arena settings ////////
	BYTE * Map;					// Map that the bot sees
	BYTE * ChecksumMap;			// Official map contents
	BYTE arena_settings[1428];
	ArenaSettings Settings;

	//////// Players ////////
	_linkedlist <Player> PlayerList;		// Linked list of players
	_referencetable <Player> PlayerQAL;		// Reference table of players
	Player * Me;

	void AddPlayer(BYTE PlayerID,
				   BYTE ShipType,
				   const char * PlayerName,
				   const char * SquadName,
				   long FlagPoints,
				   long KillPoints,
				   short Frequency,
				   short Wins,
				   short Losses)
	{
		Player * newguy = new Player;

		newguy->SetSettings(&Settings.ships[ShipType]);
		newguy->ResetShip();

		newguy->PlayerID = PlayerID;
		newguy->ShipType = ShipType;
		memcpy(newguy->Name, PlayerName, 20);
		memcpy(newguy->Squadname, SquadName, 20);

		newguy->FlagPoints	= FlagPoints;
		newguy->KillPoints	= KillPoints;
		newguy->Frequency	= Frequency;
		newguy->Wins		= Wins;
		newguy->Losses		= Losses;

		newguy->x	= 0;
		newguy->y	= 0;
		newguy->tx	= 0;
		newguy->ty	= 0;

		PlayerList.Append(newguy);
		PlayerQAL.AddItem(PlayerID, newguy);

		newguy->SetBounceProximity(10);
		newguy->SetWallFriction(Settings.BounceFactor);

		if (PlayerID == MyID)
		{
			Me = newguy;
		}
	}

	void NukePlayer(BYTE PlayerID)
	{
		Player * oldguy = PlayerQAL.Get(PlayerID);

		// Delete player
		PlayerQAL.NukeItem(PlayerID);
		PlayerList.Delete(oldguy, false);

		// Delete player references in PlayerList
		_listnode <Player> * player = PlayerList.head;
		while (player)
		{
			Player * guy = player->item;
			player = player->next;

			if (guy->AttachedTo == oldguy)
				guy->AttachedTo = NULL;
		}

		// Delete player references in WeaponList
		_listnode <Weapon> * weapon = WeaponList.head;
		while (weapon)
		{
			_listnode <Weapon> * guy = weapon;
			weapon = weapon->next;

			if (guy->item->Owner == oldguy)
				WeaponList.Delete(guy, false);
		}

		// Delete player references in SoccerList
		_listnode <Powerball> * ball = SoccerList.head;
		while (ball)
		{
			Powerball * guy = ball->item;;
			ball = ball->next;

			if (guy->Controller == oldguy)
			{
				guy->Controller = NULL;
				guy->Carried = false;;
			}
		}

		// Delete player references in FlagList
		_listnode <FlagData> * flag = FlagList.head;
		while (flag)
		{
			FlagData * guy = flag->item;;
			flag = flag->next;

			if (guy->Controller == oldguy)
			{
				guy->Controller = NULL;
				guy->Carried = false;;
			}
		}
	}

	//////// Soccer game ////////
	_linkedlist <Powerball> SoccerList;

	Powerball * FindPowerball(SHORT ID);
	void PollBallsForKickOff();

	//////// Flag game ////////
	_linkedlist <FlagData> FlagList;

	FlagData * FindFlag(SHORT ID);
	void PollFlagsForKickOff();

	//////// Weapon trajectories ////////
	_linkedlist <Weapon> WeaponList;

	void GenerateWeapon(Player * player, WeaponInfo Type);
	void PollWeaponsForKickOff();

	//////// Brick blocks ////////
	_linkedlist <WeaponImpact> ImpactList;

	//////// Brick blocks ////////
	_linkedlist <Brick> BrickList;

	void PollBricksForKickOff();

	//////// Arena macros ////////
	void SendPublicChat(const char * Message);

	//////// Sync ////////
	SHORT LastPing;				// Set by sync packet & used by security checksum packet
	SHORT AveragePing;			// " "
	SHORT LowestPing;			// " "
	SHORT HighestPing;			// " "
	DWORD DiffTime;				// Difference in local and remote 'tick counts
	DWORD TotalSent;			// Count of sent packets
	DWORD TotalRecvd;			// Count of received packets
	DWORD WeaponCount;			// Count of received weapons
	WORD S2CSlowCurrent;		// Last count of slow packets
	WORD S2CFastCurrent;		// Last count of fast packets
	DWORD S2CSlowTotal;			// Total slow packets
	DWORD S2CFastTotal;			// Total fast packets

	DWORD GetGameTime();
	DWORD GetLocalTime();
	DWORD LocalToGameTime(DWORD LocalTime);
	DWORD GameToLocalTime(DWORD GameTime);

	DWORD LastPositionUpdate;	// Time at which the last position packet was sent
	DWORD PositionUpdateCount;	// After 40, send a sync packet

	//////// Session ////////
	DWORD	LastTicks;	// Used to calculate time since the last packet was recv'd
	DWORD	Ping;		// Mean ping-time
	DWORD	CloseID;	// When this ACK_ID is recv'd, the subbill closes the session

	void SetTicks();

	//////// Encryption ////////
	SSET SSEncrTable;
	DWORD SentKey;

	void _cdecl InitializeEncryption(DWORD Key);

	//////// File access ////////
	void ParseMapFile(char * FileContents, LONG Length);
	bool LoadMap(const char * MapFilename, LONG Checksum);
	void SaveMap(const BYTE * MapFilename, BYTE * MapBytes, LONG MapLength);

	//////// Reliable packet system ////////
	DWORD						RemoteStep;	// Next packet number to-be-processed
	_linkedlist <PacketStruct>	Backlog;	// List of packets-to-be-processed

	DWORD						LocalSend;	// Last packet to be sent
	_linkedlist <PostlogNode>	Sendlog;	// List of packets being sent

	DWORD						LocalPost;	// Next packet number queued for delivery
	_linkedlist <PostlogNode>	Postlog;	// List of packets queued for sending

	void SendReliable(BYTE * Message, DWORD Length);	// Note: Do not use this as
														// a substitute for Send();
	void CheckPostlog(DWORD PacketID);
	bool CheckBacklog(PacketStruct * Packet);

	//////// Car pooling ////////
	bool Clustering;				// Are we clustering?
	_linkedlist <LoggedPacket> ClusterMessages;
	long TotalReliableClusters;

	void StartCluster();
	void EndCluster();

	void Send(BYTE * Message, DWORD Length, bool Reliable);		// Hook for clustering
	void SendTo(BYTE * Message, DWORD Length, bool Reliable);	// Actually sends packets

	//////// Chunks ////////
	IncomingChunk IncomingChunks;	// 08 - 09
	IncomingChunk BIGChunks;		// 0a <Length>

	//////// Handlers ////////
	_jumptable <PacketStruct> Router;
	_jumptable <PacketStruct> SpecialRouter;

public:
	//////// Touch these and you will die a horrible, slow and painful death ////////
	INADDR Remote;					// Host address
	ClientSettings INISettings;		// Settings from the INI file

	//////// Encryption ////////
	void Encrypt(BYTE * Message, DWORD Length);
	void Decrypt(BYTE * Message, DWORD Length);

	//////// Messages ////////
	void SetHandler(BYTE Type, bool (*function)(PacketStruct *));
	void SetSpecialHandler(BYTE Type, bool (*function)(PacketStruct *));
	void ClearHandlers();

	//////// Session params ////////
	void UpdatePing(DWORD Ping);
	LONG SetKey(LONG Key);

	//////// Socket data ////////
	SOCKET SocketID;

	//////// Socket operations ////////
	void Route(BYTE * Message, LONG Length);
	void Disconnect(bool Wait);
	bool ResendQueue();

	//////// Game physics ////////
	void SendPosition();

	//////// System parameters ////////
	void RetrieveSystemInformation();
	bool ValidateNewsChecksum(DWORD Checksum);
	void AddNewsChecksum(DWORD Checksum);

	//////// Initialization ////////
	bool Running;

	void Initialize();
	void Connect(DWORD IP, SHORT Port);
	void Terminate();
	void ResetSession()
	{
		INISettings.ReloadSettings();

		SendingPosition		= false;
		PasswordWait		= false;
		HasSession			= false;
		SSEncrTable.SSKey	= 0;
		RemoteStep			= 0;
		LocalSend			= 0;
		LocalPost			= 0;
		CloseID				= 0xFFFFFFFF;		// extend this time somehow
		Ping				= 500;
		SyncTime			= 0;
		LastPositionUpdate	= GetTickCount();
		PositionUpdateCount	= 0;
		LastTicks			= GetTickCount();
		Me					= NULL;
		TotalSent			= 1;
		TotalRecvd			= 0;
		WeaponCount			= 0;
		S2CSlowCurrent		= 0;
		S2CFastCurrent		= 0;
		S2CSlowTotal		= 0;
		S2CFastTotal		= 0;

		LastPing			= 0;
		AveragePing			= 0;
		LowestPing			= 0;
		HighestPing			= 0;

		for (int i = 0; i < 256; i++)
		{
			SetSpecialHandler(i,	HandleUnknownSpecial);
			SetHandler(i,			HandleUnknown);
		}

		SetHandler(0,				HandleSpecial);

		SetSpecialHandler(2,		HandleEncryptResponse);
		SetSpecialHandler(5,		HandleSyncRequest);
		SetSpecialHandler(7,		HandleDisconnect);

		if (Map)
		{
			delete Map;
			delete ChecksumMap;
		}

		Map = (BYTE*) calloc(1024, 1024);
		memset(Map, 0, 1024 * 1024);

		ChecksumMap = (BYTE*) calloc(1024, 1024);
		memset(ChecksumMap, 0, 1024 * 1024);

		Radar.SetMap(Map);

		RetrieveSystemInformation();
	}

	BotSession() : Mutex(),
				   Chat(&Window, 0, 480 - 320, 800, 320),
				   Font(&Window,		"gfx/font.bmp"),
				   ShipSet(&Window,		"gfx/ships.bmp",	36, 36),
				   Asteroids(&Window,	"gfx/asteroid1.bmp",16, 16),
				   Asteroids2(&Window,	"gfx/asteroid2.bmp",32, 32),
				   Station(&Window,		"gfx/station.bmp",	96, 96),
				   Wormhole(&Window,	"gfx/wormhole.bmp",	80, 80),
				   TileSet(&Window,		"gfx/tiles.bmp",	16, 16),
				   BombSet(&Window,		"gfx/bombs.bmp",	16, 16),
				   BulletSet(&Window,	"gfx/bullets.bmp",	5, 5),
				   FlagSet(&Window,		"gfx/flags.bmp",	16, 16),
				   MineSet(&Window,		"gfx/mines.bmp",	16, 16),
				   BrickSet(&Window,	"gfx/bricks.bmp",	16, 16),
				   ShrapSet(&Window,	"gfx/shrap.bmp",	8, 8),
				   BallSet(&Window,		"gfx/ball.bmp",		16, 16),
				   GoalSet(&Window,		"gfx/goals.bmp",	16, 16),
				   RepelSet(&Window,	"gfx/repel.bmp",	96, 96),
				   GreenSet(&Window,	"gfx/greens.bmp",	16, 16),
				   TurretSet(&Window,	"gfx/turret.bmp",	40, 40),
				   StatusBox(&Window,	"gfx/statusbox.bmp",84, 68),
				   NumberFont1(&Window,	"gfx/statfont1.bmp",16, 24),
				   NumberFont2(&Window,	"gfx/statfont2.bmp",7, 10),
				   NumberSet1(&Window, true),
				   NumberSet2(&Window, true),
				   ItemFont(&Window,	"gfx/itemfont.bmp",	11, 13),
				   ItemSet(&Window,		"gfx/items.bmp",	26, 24),
				   ExplosionSet(&Window,"gfx/bombimpact.bmp",	48, 48),
				   ImpactSet(&Window,	"gfx/bulletimpact.bmp",	16, 16),
				   ChatBox(&Window, 100),
				   Radar(&Window),
				   Starfield(&Window)
	{
#ifdef TRACE_CONSTRUCTORS
	MessageBox(NULL, "In BotSession() constructor", "Loading...", 0);
#endif

		Running		= true;
		Clustering	= false;

		ThreadsOpen = 0;

		hResend = hGraphics = hReceive = 0;

		Chat.SetFont(&Font);
		ChatBox.SetFont(&Font);

		NumberSet1.SetFont(&NumberFont1);
		NumberSet2.SetFont(&NumberFont2);

		MapWidth		= (1024 << 4);
		MapHeight		= (1024 << 4);
		AnimationFrame	= 0;
		Map				= NULL;

#ifdef TRACE_CONSTRUCTORS
	MessageBox(NULL, "Out of BotSession() constructor", "Loading...", 0);
#endif
	}

	~BotSession()
	{
		if (hResend || hReceive || hGraphics)
			Terminate();

		if (Map)
		{
			delete Map;
			delete ChecksumMap;
		}
	}
};


// Implementations

#include "bot.h"
#include "encryption.h"
#include "rpc.h"
#include "command.h"
